"""
This module contains default style constants for visualization methods in starfish
"""

import matplotlib.pyplot as plt

TARGETS_COLORMAP = plt.get_cmap('nipy_spectral')

BACKGROUND_COLORMAP = plt.get_cmap('gray')

# TODO ambrosejcarr: propagate this through plots
FONTSIZE = 14
