from typing import Union

from ._constants import (
    Coordinates,
    Features,
    Indices,
    PHYSICAL_COORDINATE_DIMENSION,
    PhysicalCoordinateTypes,
)
from ._spot_attributes import SpotAttributes

Number = Union[int, float]
